from torchline.utils import Registry

META_ARCH_REGISTRY = Registry("META_ARCH")
META_ARCH_REGISTRY.__doc__ = """
Registry for meta-architectures, i.e. the whole model.

The registered object will be called with `obj(cfg)`
and expected to return a `nn.Module` object.
"""

__all__ = [
    'build_model',
    'META_ARCH_REGISTRY'
]

def build_model(cfg):
    """
    Built the whole model, defined by `cfg.model.name`.
    """
    name = cfg.model.name
    return META_ARCH_REGISTRY.get(name)(cfg)
    


import pretrainedmodels
import torch.nn as nn

from torchline.models import META_ARCH_REGISTRY

__all__ = [
    'Nasnetamobile',
    'Pnasnet5large',
]

def generate_model(cfg, name):
    pretrained='imagenet' if cfg.model.pretrained else None
    classes = cfg.model.classes
    img_size = cfg.input.size[0]
    model = pretrainedmodels.__dict__[name](num_classes=1000, pretrained=pretrained)
    model.avg_pool = nn.AdaptiveAvgPool2d(1)
    if classes != 1000:
        in_features = model.last_linear.in_features
        model.last_linear = nn.Sequential(
            nn.Linear(in_features, classes, bias=False))
    return model

@META_ARCH_REGISTRY.register()
def Nasnetamobile(cfg):
    return generate_model(cfg, 'nasnetamobile')

@META_ARCH_REGISTRY.register()
def Pnasnet5large(cfg):
    return generate_model(cfg, 'pnasnet5large')



import torch

__all__ = [
    'generate_optimizer',
    'generate_scheduler'
]

def generate_optimizer(model, optim_name, lr, momentum=0.9, weight_decay=1e-5):
    '''
    return torch.optim.Optimizer
    '''
    if optim_name.lower() == 'sgd':
        return torch.optim.SGD(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
    elif optim_name.lower() == 'adadelta':
        return torch.optim.Adagrad(model.parameters(), lr=lr, weight_decay=weight_decay)
    elif optim_name.lower() == 'adam': 
        return torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
    elif optim_name.lower() == 'rmsprop':
        return torch.optim.RMSprop(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
    else:
        print(f"{optim_name} not implemented")
        raise NotImplementedError

def generate_scheduler(optimizer, scheduler_name, **params):
    '''
    return torch.optim.lr_scheduler
    '''
    if scheduler_name.lower() == 'CosineAnnealingLR'.lower():
        return torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, **params)
    elif scheduler_name.lower() == 'CosineAnnealingWarmRestarts'.lower():
        return torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, **params)
    elif scheduler_name.lower() == 'StepLR'.lower():
        return torch.optim.lr_scheduler.StepLR(optimizer, **params)
    elif scheduler_name.lower() == 'MultiStepLR'.lower():
        return torch.optim.lr_scheduler.MultiStepLR(optimizer, **params)
    elif scheduler_name.lower() == 'ReduceLROnPlateau'.lower():
        return torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, **params)
    else:
        print(f"{scheduler_name} not implemented")
        raise NotImplementedError


import glob
import os

import torch
from PIL import Image
from torchvision import transforms

__all__ = [
    'image_loader',
    'get_imgs_to_predict',
    'topk_acc',
    'model_size'
]

def image_loader(filename, cfg):
    '''load an image and convert it to tensor
    Args:
        filename: image filename
        cfg: CfgNode
    return:
        torch.tensor
    '''
    image = Image.open(filename).convert('RGB')
    mean = cfg.transforms.tensor.normalization.mean
    std = cfg.transforms.tensor.normalization.std
    img_size = cfg.input.size

    transform = transforms.Compose([
                transforms.Resize(img_size),
                transforms.ToTensor(),
                transforms.Normalize(mean, std),
            ])
    image = transform(image)
    return image

def get_imgs_to_predict(path, cfg):
    ''''load images which are only used for prediction or testing
    Args:
        path: str
    return:
        torch.tensor (N*C*H*W)
    '''
    if os.path.isfile(path):
        images = image_loader(path, cfg).unsqueeze(0)
    elif os.path.isdir(path):
        image_types = [os.path.join(path,'*.jpg'), os.path.join(path,'*.png')]
        image_files = []
        images = {
            'img_file': [],
            'img_data': []
        }
        for img_type in image_types:
            image_files.extend(glob.glob(img_type))
        for img_file in image_files:
            images['img_file'].append(img_file)
            images['img_data'].append(image_loader(img_file, cfg))
        images['img_data'] = torch.stack(images['img_data'])
    return images

def model_size(model):
    return sum([p.numel() for p in model.parameters()])*4/1024**2

def topk_acc(output, target, topk=(1, 3)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k*100.0/len(target.view(-1)))
        # res.append(correct_k.mul_(100.0 / batch_size))
    return torch.tensor(res)



from albumentations import (CLAHE, Blur, ChannelDropout, ChannelShuffle,
                            Compose, Cutout, Flip, GaussNoise, GridDistortion,
                            HueSaturationValue, IAAAdditiveGaussianNoise,
                            IAAEmboss, MotionBlur, Normalize, OneOf,
                            OpticalDistortion, RandomBrightnessContrast,
                            RandomGridShuffle, Resize, ShiftScaleRotate)
from albumentations.pytorch.transforms import ToTensor

from .transforms import TRANSFORMS_REGISTRY, BaseTransforms

__all__ = [
    'AlbumentationsTransforms'
]

@TRANSFORMS_REGISTRY.register()
def AlbumentationsTransforms(cfg):
    is_train = cfg.dataset.is_train
    log_name = cfg.log.name
    mean = cfg.transforms.tensor.normalization.mean
    std = cfg.transforms.tensor.normalization.std
    img_size = cfg.input.size
    random_grid_shuffle = cfg.abtfs.random_grid_shuffle
    channel_shuffle = cfg.abtfs.channel_shuffle
    channel_dropout = cfg.abtfs.channel_dropout
    noise = cfg.abtfs.noise
    blur = cfg.abtfs.blur
    distortion = cfg.abtfs.distortion
    bright = cfg.abtfs.bright
    hue = cfg.abtfs.hue
    cutout = cfg.abtfs.cutout
    rotate = cfg.abtfs.rotate
    return _AlbumentationsTransforms(is_train, log_name, img_size, 
            random_grid_shuffle,
            channel_shuffle,
            channel_dropout,
            noise,
            blur,
            distortion,
            bright,
            hue,
            cutout,
            rotate,
            mean, std)

class _AlbumentationsTransforms(BaseTransforms):
    def __init__(self, is_train, log_name, img_size,
                 random_grid_shuffle={'enable':0, 'grid':3},
                 channel_shuffle={'enable':0, 'p':1},
                 channel_dropout={'enable':0, 'drop_range':(1, 1), 'fill_value':127},
                 noise={'enable':0, 'p':1},
                 blur={'enable':0, 'p':1, 'blur_limit':3},
                 distortion={'enable':0, 'p':1},
                 bright={'enable':0, 'p':1, 'clip_limit':2},
                 hue={'enable':0, 'p':1},
                 cutout={'enable':0, 'p':1, 'size':16, 'fill_value':127},
                 rotate={'enable':0, 'p':1, 'shift_limit': 0.0625, 'scale_limit':0.2, 'rotate_limit':45},
                 mean=[0.5,0.5,0.5], std=[0.5,0.5,0.5]):
        super(_AlbumentationsTransforms, self).__init__(is_train, log_name)
        self.is_train = is_train
        self.img_size = img_size
        self.mean = mean
        self.std = std
        self.normalize = Normalize(self.mean, self.std)
        self.random_grid_shuffle = random_grid_shuffle
        self.channel_shuffle = channel_shuffle
        self.channel_dropout = channel_dropout
        self.noise = noise
        self.blur = blur
        self.distortion = distortion
        self.bright = bright
        self.hue = hue
        self.cutout = cutout
        self.rotate = rotate
        self.init_transforms = self.get_transform()
        self.transform = self.parse_transform

    def parse_transform(self, image):
        image = self.init_transforms(image=image)['image']
        return image

    @property
    def valid_transform(self):
        height, width = self.img_size
        return Compose([Resize(height, width),
            self.normalize,
            ToTensor()])

    @property
    def train_transform(self):
        height, width = self.img_size
        transforms_list = [Resize(height, width), Flip()]
        
        # random_grid_shuffle
        if self.random_grid_shuffle['enable']:
            grid = self.random_grid_shuffle['grid']
            grid = (grid,grid)
            transforms_list.append(RandomGridShuffle((grid)))

        # channel_shuffle
        if self.channel_shuffle['enable']:
            transforms_list.append(ChannelShuffle(p=1))
        
        # channel_dropout
        if self.channel_dropout['enable']:
            drop_range = self.channel_dropout.drop_range
            fill_value = self.channel_dropout.fill_value
            transforms_list.append(ChannelDropout(drop_range, fill_value, p=1))

        # noise
        if self.noise['enable']:
            transforms_list.append(OneOf([
                IAAAdditiveGaussianNoise(),
                GaussNoise(),
            ], p=1))

        # blur
        if self.blur['enable']:
            transforms_list.append(OneOf([
                MotionBlur(),
                Blur(blur_limit=3,),
            ], p=1))

        # rotate
        if self.rotate['enable']:
            params = {key:value for key, value in self.rotate.items() if key != 'enable' }
            transforms_list.append(ShiftScaleRotate(**params))

        # distortion
        if self.distortion['enable']:
            transforms_list.append(OneOf([
                OpticalDistortion(p=0.3),
                GridDistortion(p=.3),
            ], p=1))

        # bright
        if self.bright['enable']:
            transforms_list.append(
                OneOf([
                    CLAHE(clip_limit=self.bright['clip_limit']),
                    RandomBrightnessContrast(p=0.8),            
                ], p=1))

        # hue color
        if self.hue['enable']:
            transforms_list.append(HueSaturationValue(p=0.3))
        
        # cutout
        if self.cutout['enable']:
            num_holes = self.cutout['num_holes']
            size = self.cutout['size']
            fill_value = self.cutout['fill_value']
            transforms_list.append(Cutout(num_holes, size, size, fill_value, 1))
        transforms_list.append(self.normalize)
        transforms_list.append(ToTensor())

        return Compose(transforms_list)



import torch
from torchvision.datasets import CIFAR10 as _CIFAR10
from torchvision.datasets import MNIST as _MNIST

from .build import DATASET_REGISTRY
from .transforms import build_transforms

__all__ = [
    'MNIST',
    'CIFAR10',
    'FakeData',
    'fakedata'
]

@DATASET_REGISTRY.register()
def MNIST(cfg):
    root = cfg.dataset.dir
    is_train = cfg.dataset.is_train
    transform = build_transforms(cfg)
    return _MNIST(root=root, train=is_train, transform=transform.transform, download=True)

@DATASET_REGISTRY.register()
def CIFAR10(cfg):
    root = cfg.dataset.dir
    is_train = cfg.dataset.is_train
    transform = build_transforms(cfg)
    return _CIFAR10(root=root, train=is_train, transform=transform.transform, download=True)

class FakeData(torch.utils.data.Dataset):
    def __init__(self, size=64, num=100):
        if isinstance(size, int):
            self.size = [size, size]
        elif isinstance(size, list):
            self.size = size
        self.num = num
        self.data = torch.rand(num, 3, *size)
        self.labels = torch.randint(0, 10, (num,))


    def __getitem__(self, index):
        return self.data[index], self.labels[index]

    def __len__(self):
        return self.num

@DATASET_REGISTRY.register()
def fakedata(cfg):
    size = cfg.input.size
    return FakeData(size)


from .build import *
from .resnet import *
from .pnasnet import *
from .efficientnet import *
from .dpn import *
from .nas_models import *
from .mnasnet import *
from .mobilenet import *

model_list = [
    'DPN26',
    'DPN92',
    'EfficientNetB0',
    'MNASNet0_5',
    'MNASNet0_75',
    'MNASNet1_0',
    'MNASNet1_3',
    'MobileNet_V2',
    'Nasnetamobile',
    'Pnasnet5large',
    'PNASNetA',
    'PNASNetB',
    'Resnet18',
    'Resnet34',
    'Resnet50',
    'Resnet101',
    'Resnet152',
    'Resnext50_32x4d',
    'Resnext101_32x8d',
    'Wide_resnet50_2',
    'Wide_resnet101_2',
]



from torchline.utils import Registry, Logger

MODULE_REGISTRY = Registry('MODULE')
MODULE_REGISTRY.__doc__ = """
Registry for module template, e.g. DefaultModule.

The registered object will be called with `obj(cfg)`
"""

__all__ = [
    'MODULE_REGISTRY',
    'build_module'
]

def build_module(cfg):
    """
    Built the module template, defined by `cfg.module.name`.
    """
    name = cfg.module.name
    return MODULE_REGISTRY.get(name)(cfg)

    


import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.mobilenet import MobileNetV2

from .build import META_ARCH_REGISTRY

__all__ = [
    'MobileNetV2',
    'MobileNet_V2'
]



class MobileNetV2(nn.Module):
    # Modify attributs	    
    def __init__(self, model):
        super(MobileNetV2, self).__init__()
        for key, val in model.__dict__.items():
            self.__dict__[key] = val

        self.g_avg_pool = nn.AdaptiveAvgPool2d(1)
        self.last_linear = self.classifier

    def logits(self, x):
        out = self.g_avg_pool(x)
        out = out.view(out.size(0), -1)
        out = self.last_linear(out)
        return out

    def forward(self, x):
        out = self.features(x)
        out = self.logits(out)
        return out

def generate_model(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    if 'dropout' in cfg.model:
        dropout = cfg.model.dropout
    else:
        dropout = 0.2
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.classifier[1].in_features
        model.classifier = nn.Sequential(
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(in_features, classes, bias=False))
    return MobileNetV2(model)

@META_ARCH_REGISTRY.register()
def MobileNet_V2(cfg):
    return generate_model(cfg, 'mobilenet_v2')


import random
from .config import CfgNode as CN

_C = CN()
_C.VERSION = 1


# ---------------------------------------------------------------------------- #
# input
# ---------------------------------------------------------------------------- #
_C.input = CN()
_C.input.size = (224, 224)

# ---------------------------------------------------------------------------- #
# dataset
# ---------------------------------------------------------------------------- #
_C.dataset = CN()
_C.dataset.name = 'cifar10'
_C.dataset.batch_size = 16
_C.dataset.dir = './datasets/skin100_dataset/train'
_C.dataset.train_list = './datasets/train_skin10.txt'
_C.dataset.valid_list = './datasets/valid_skin10.txt'
_C.dataset.test_list = './datasets/test_skin10.txt'
_C.dataset.is_train = False # specify to load training or testing set


# ---------------------------------------------------------------------------- #
# transforms
# ---------------------------------------------------------------------------- #

_C.transforms = CN() # image transforms
_C.transforms.name = 'DefaultTransforms'


## transforms for tensor
_C.transforms.tensor = CN()
# for skin100
_C.transforms.tensor.normalization = CN()
_C.transforms.tensor.normalization.mean = [0.6075, 0.4564, 0.4182] 
_C.transforms.tensor.normalization.std = [0.2158, 0.1871, 0.1826]
# _C.transforms.tensor.normalization = {
#             'mean':[0.6054, 0.4433, 0.4084], 
#             'std': [0.2125, 0.1816, 0.1786]  # for skin10
_C.transforms.tensor.random_erasing = CN()
_C.transforms.tensor.random_erasing.enable = 0
_C.transforms.tensor.random_erasing.p = 0.5
_C.transforms.tensor.random_erasing.scale = (0.02, 0.3) # range of proportion of erased area against input image.
_C.transforms.tensor.random_erasing.ratio = (0.3, 3.3), # range of aspect ratio of erased area.


# ---------------------------------------------------------------------------- #
# albumentations transforms (abtfs)
# ---------------------------------------------------------------------------- #

_C.abtfs = CN()
_C.abtfs.random_grid_shuffle = CN()
_C.abtfs.random_grid_shuffle.enable = 0
_C.abtfs.random_grid_shuffle.grid = 2

_C.abtfs.channel_shuffle = CN()
_C.abtfs.channel_shuffle.enable = 0

_C.abtfs.channel_dropout = CN()
_C.abtfs.channel_dropout.enable = 0
_C.abtfs.channel_dropout.drop_range = (1, 1)
_C.abtfs.channel_dropout.fill_value = 127

_C.abtfs.noise = CN()
_C.abtfs.noise.enable = 1

_C.abtfs.blur = CN()
_C.abtfs.blur.enable = 0
_C.abtfs.blur.blur_limit = 3

_C.abtfs.rotate = CN()
_C.abtfs.rotate.enable = 1
_C.abtfs.rotate.p = 1
_C.abtfs.rotate.shift_limit = 0.0625
_C.abtfs.rotate.scale_limit = 0.2
_C.abtfs.rotate.rotate_limit = 45
                 
_C.abtfs.bright = CN()
_C.abtfs.bright.enable = 1
_C.abtfs.bright.clip_limit = 1

_C.abtfs.distortion = CN()
_C.abtfs.distortion.enable = 0

_C.abtfs.hue = CN()
_C.abtfs.hue.enable = 0

_C.abtfs.cutout = CN()
_C.abtfs.cutout.enable = 1
_C.abtfs.cutout.num_holes = 10
_C.abtfs.cutout.size = 20
_C.abtfs.cutout.fill_value = 127

# ---------------------------------------------------------------------------- #
# torchvision transforms
# ---------------------------------------------------------------------------- #

## transforms for PIL image
_C.transforms.img = CN()

### modify the image size, only use one operation
# random_resized_crop
_C.transforms.img.random_resized_crop = CN()
_C.transforms.img.random_resized_crop.enable = 0
_C.transforms.img.random_resized_crop.scale = (0.5, 1.0)
_C.transforms.img.random_resized_crop.ratio = (3/4, 4/3)

# resize
_C.transforms.img.resize =  CN()
_C.transforms.img.resize.enable = 1

# random_crop
_C.transforms.img.random_crop = CN()
_C.transforms.img.random_crop.enable = 1
_C.transforms.img.random_crop.padding = 0

# center_crop
_C.transforms.img.center_crop = CN()
_C.transforms.img.center_crop.enable = 0

### without modifying the image size
_C.transforms.img.aug_imagenet = False
_C.transforms.img.aug_cifar = False

# color_jitter
_C.transforms.img.color_jitter = CN()
_C.transforms.img.color_jitter.enable = 0
_C.transforms.img.color_jitter.brightness = 0.
_C.transforms.img.color_jitter.contrast = 0.
_C.transforms.img.color_jitter.saturation = 0.
_C.transforms.img.color_jitter.hue = 0.

# horizontal_flip
_C.transforms.img.random_horizontal_flip = CN()
_C.transforms.img.random_horizontal_flip.enable = 1
_C.transforms.img.random_horizontal_flip.p = 0.5

# vertical_flip
_C.transforms.img.random_vertical_flip = CN()
_C.transforms.img.random_vertical_flip.enable = 1
_C.transforms.img.random_vertical_flip.p = 0.5

# random_rotation
_C.transforms.img.random_rotation = CN()
_C.transforms.img.random_rotation.enable = 1
_C.transforms.img.random_rotation.degrees = 10



_C.label_transforms = CN() # label transforms
_C.label_transforms.name = 'default'


# ---------------------------------------------------------------------------- #
# dataloader
# ---------------------------------------------------------------------------- #
_C.dataloader = CN()
_C.dataloader.num_workers = 4
_C.dataloader.sample_train = "default"
_C.dataloader.sample_test = "default"


# ---------------------------------------------------------------------------- #
# model
# ---------------------------------------------------------------------------- #
_C.model = CN()
_C.model.name = 'Resnet50'
_C.model.classes = 10
_C.model.pretrained = True
_C.model.finetune = False


# ---------------------------------------------------------------------------- #
# optimizer
# ---------------------------------------------------------------------------- #
_C.optim = CN()
_C.optim.name = 'adam'
_C.optim.momentum = 0.9
_C.optim.base_lr = 0.001
# _C.optim.lr = _C.optim.base_lr # will changed in v0.3.0.0
_C.optim.weight_decay = 0.0005

# scheduler
_C.optim.scheduler = CN()
_C.optim.scheduler.name = 'MultiStepLR'
_C.optim.scheduler.gamma = 0.1 # decay factor

# for CosineAnnealingLR
_C.optim.scheduler.t_max = 10 

# for CosineAnnealingLR
_C.optim.scheduler.t_0 = 5
_C.optim.scheduler.t_mul = 20

# for ReduceLROnPlateau
_C.optim.scheduler.mode = 'min' # min for loss, max for acc
_C.optim.scheduler.patience = 10
_C.optim.scheduler.verbose = True # print log once update lr

# for StepLR
_C.optim.scheduler.step_size = 10

# for MultiStepLR
_C.optim.scheduler.milestones = [10, 25, 35, 50]

# _C.optimizer = _C.optim # enhance compatibility. will changed in v0.3.0.0
# ---------------------------------------------------------------------------- #
# loss
# ---------------------------------------------------------------------------- #
_C.loss = CN()
_C.loss.name = 'CrossEntropy'
_C.loss.class_weight = []
_C.loss.label_smoothing = 0.1 # CrossEntropyLabelSmooth

_C.loss.focal_loss = CN()
_C.loss.focal_loss.alpha = [] # FocalLoss
_C.loss.focal_loss.gamma = 2
_C.loss.focal_loss.size_average = True
# ---------------------------------------------------------------------------- #
# hooks
# ---------------------------------------------------------------------------- #
_C.hooks = CN()

## EarlyStopping
_C.hooks.early_stopping = CN()
_C.hooks.early_stopping.setting = 2 # 0: True 1: False 2: custom
_C.hooks.early_stopping.monitor = 'valid_loss'
_C.hooks.early_stopping.min_delta = 0.
_C.hooks.early_stopping.patience = 10
_C.hooks.early_stopping.mode = 'min'
_C.hooks.early_stopping.verbose = 1

# ModelCheckpoint
_C.hooks.model_checkpoint = CN()
_C.hooks.model_checkpoint.setting = 0 # 0: True 1: False 2: custom
_C.hooks.model_checkpoint.filepath = '' # the empty file path is recommended
_C.hooks.model_checkpoint.monitor = 'valid_loss'
_C.hooks.model_checkpoint.mode = 'min'
_C.hooks.model_checkpoint.verbose = 1


# ---------------------------------------------------------------------------- #
# Module template 
# ---------------------------------------------------------------------------- #

_C.module = CN()
_C.module.name = 'DefaultModule'

# ---------------------------------------------------------------------------- #
# Trainer 
# ---------------------------------------------------------------------------- #

_C.trainer = CN()
_C.trainer.name = 'DefaultTrainer'
_C.trainer.default_save_path = './output' # will be removed
_C.trainer.default_root_dir = './output'
_C.trainer.gradient_clip_val = 0 # 0 means don't clip.
_C.trainer.process_position = 0
_C.trainer.num_nodes = 1
_C.trainer.gpus = [] # list
_C.trainer.auto_select_gpus = False # If `auto_select_gpus` is enabled and `gpus` is an integer, pick available gpus automatically. 
                                    # This is especially useful when GPUs are configured to be  in "exclusive mode", such that 
                                    # only one process at a time can access them.
_C.trainer.num_tpu_cores = '' # How many TPU cores to train on (1 or 8).
_C.trainer.log_gpu_memory = "" # None, 'min_max', 'all'. Might slow performance
_C.trainer.show_progress_bar = False # will be removed
_C.trainer.progress_bar_refresh_rate = 0 # How often to refresh progress bar (in steps). Value ``0`` disables progress bar.
_C.trainer.overfit_pct = 0.0 # if 0<overfit_pct<1, (e.g. overfit_pct = 0.1) then train, val, test only 10% data.
_C.trainer.track_grad_norm = -1 # -1 no tracking. Otherwise tracks that norm. if equals to 2, then 2-norm will be traced
_C.trainer.check_val_every_n_epoch = 1 # Check val every n train epochs.
_C.trainer.fast_dev_run = False # everything only with 1 training and 1 validation batch.
_C.trainer.accumulate_grad_batches = 1 # Accumulates grads every k batches or as set up in the dict.
_C.trainer.max_epochs = 200 # Stop training once this number of epochs is reached.
_C.trainer.min_epochs = 100 # Force training for at least these many epochs
_C.trainer.max_steps = 100 # Stop training after this number of steps. Disabled by default (None).
_C.trainer.min_steps = 100 # Force training for at least these number of steps. Disabled by default (None).
_C.trainer.train_percent_check = 1.0 # How much of training dataset to check.
_C.trainer.val_percent_check = 1.0 # How much of validation dataset to check.
_C.trainer.test_percent_check = 1.0 # How much of test dataset to check.
_C.trainer.val_check_interval = 1.0 # How often within one training epoch to check the validation set
_C.trainer.log_save_interval = 100 # Writes logs to disk this often
_C.trainer.row_log_interval = 10 # How often to add logging rows (does not write to disk)
_C.trainer.distributed_backend = 'dp' # 'dp', 'ddp', 'ddp2', 'horovod'
_C.trainer.use_amp = False # will be removed, please use `precision` as below instead.
_C.trainer.amp_level = 'O1' # will be removed, please use `precision` as below instead.
_C.trainer.precision = 32 # Full precision (32), half precision (16).
_C.trainer.print_nan_grads = True # Prints gradients with nan values. will be removed
_C.trainer.weights_summary = '' # 'full', 'top', None.
_C.trainer.weights_save_path = '' # Where to save weights if specified. Will override default_root_dir for checkpoints only. 
                                  # Use this if for whatever reason you need the checkpoints stored in a different place 
                                  # than the logs written in `default_root_dir`.
_C.trainer.num_sanity_val_steps = 5 # Sanity check runs n batches of val before starting the training routine.
_C.trainer.truncated_bptt_steps = ''
_C.trainer.resume_from_checkpoint = ''
# _C.trainer.profiler = ''
_C.trainer.reload_dataloaders_every_epoch = False # Set to True to reload dataloaders every epoch
_C.trainer.auto_lr_find = False # If set to True, will `initially` run a learning rate finder,
                                # trying to optimize initial learning for faster convergence. Sets learning
                                # rate in self.hparams.lr | self.hparams.learning_rate in the lightning module.
                                # To use a different key, set a string instead of True with the key name.

_C.trainer.replace_sampler_ddp = True # Explicitly enables or disables sampler replacement. 
                                      # If not specified this will toggled automatically ddp is used

_C.trainer.benchmark = False # If true enables cudnn.benchmark.
_C.trainer.terminate_on_nan = False# If set to True, will terminate training (by raising a `ValueError`) at the
    # end of each training batch, if any of the parameters or the loss are NaN or +/-inf.

_C.trainer.auto_scale_batch_size = False # If set to True, will `initially` run a batch size
                                        # finder trying to find the largest batch size that fits into memory.
                                        # The result will be stored in self.hparams.batch_size in the LightningModule.
                                        # Additionally, can be set to either `power` that estimates the batch size through
                                        # a power search or `binsearch` that estimates the batch size through a binary search.

_C.trainer.logger = CN()
_C.trainer.logger.type = 'test_tube'
_C.trainer.logger.setting = 0 # 0: True  1: False  2: custom
_C.trainer.logger.mlflow = CN()
_C.trainer.logger.mlflow.experiment_name = 'torchline_logs'
_C.trainer.logger.mlflow.tracking_uri = _C.trainer.default_root_dir
_C.trainer.logger.test_tube = CN()
_C.trainer.logger.test_tube.name = 'torchline_logs'
_C.trainer.logger.test_tube.save_dir = _C.trainer.default_root_dir
_C.trainer.logger.test_tube.version = -1 #  # if <0, then use default version. Otherwise, it will restore the version.


# ---------------------------------------------------------------------------- #
# log
# ---------------------------------------------------------------------------- #

_C.log = CN()
_C.log.path = ''
_C.log.name = 'log.txt'

# ---------------------------------------------------------------------------- #
# Misc 
# ---------------------------------------------------------------------------- #

_C.SEED = random.randint(0, 10000)
_C.DEFAULT_CUDNN_BENCHMARK = True

_C.topk = [1, 3] # save the top k results., e.g. acc@1 and acc@3

_C.predict_only = CN()
_C.predict_only.type = 'ckpt'
_C.predict_only.to_pred_file_path = '' # specify the path of images

_C.predict_only.load_ckpt = CN() # load from checkpoint
_C.predict_only.load_ckpt.checkpoint_path = '' # load_from_checkpoint

_C.predict_only.load_metric = CN()
_C.predict_only.load_metric.weights_path = '' # load_from_metrics
_C.predict_only.load_metric.tags_csv = ''
_C.predict_only.load_metric.on_gpu = True
_C.predict_only.load_metric.map_location = 'cuda:0'




from collections import OrderedDict

__all__ = [
    'AverageMeterGroup',
    'AverageMeter'
]

class AverageMeterGroup:
    """
    Average meter group for multiple average meters.
    """

    def __init__(self, verbose_type='avg'):
        self.meters = OrderedDict()
        self.verbose_type = verbose_type

    def update(self, data):
        """
        Update the meter group with a dict of metrics.
        Non-exist average meters will be automatically created.
        """
        for k, v in data.items():
            if k not in self.meters:
                self.meters[k] = AverageMeter(k, ":.4f", self.verbose_type)
            self.meters[k].update(v)

    def __getattr__(self, item):
        return self.meters[item]

    def __getitem__(self, item):
        return self.meters[item]

    def __str__(self):
        return "  ".join(f"{v}" for v in self.meters.values())

    def summary(self):
        """
        Return a summary string of group data.
        """
        return "  ".join(v.summary() for v in self.meters.values())


class AverageMeter:
    """
    Computes and stores the average and current value.
    Parameters
    ----------
    name : str
        Name to display.
    fmt : str
        Format string to print the values.
    verbose_type : str
        'all': value(avg)
        'avg': avg
    """

    def __init__(self, name, fmt=':f', verbose_type='avg'):
        self.name = name
        self.fmt = fmt
        if verbose_type not in ['all', 'avg']:
            print('Not supported verbose type, using default verbose, "avg"')
            verbose_type = 'avg'
        self.verbose_type = verbose_type
        self.reset()

    def reset(self):
        """
        Reset the meter.
        """
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """
        Update with value and weight.
        Parameters
        ----------
        val : float or int
            The new value to be accounted in.
        n : int
            The weight of the new value.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        if self.verbose_type=='all':
            fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        elif self.verbose_type=='avg':
            fmtstr = '{name} {avg' + self.fmt + '}'
        else:
            fmtstr = '{name} {avg' + self.fmt + '}'
        return fmtstr.format(**self.__dict__)

    def summary(self):
        fmtstr = '{name}: {avg' + self.fmt + '}'
        return fmtstr.format(**self.__dict__)


#coding=utf-8
import torch
import torch.nn.functional as F
import numpy as np
from .build import LOSS_FN_REGISTRY

__all__ = [
    'FocalLoss',
    '_FocalLoss'
]

@LOSS_FN_REGISTRY.register()
def FocalLoss(cfg):
    alpha = cfg.loss.focal_loss.alpha
    gamma = cfg.loss.focal_loss.gamma
    size_average = cfg.loss.focal_loss.size_average
    num_classes = cfg.model.classes
    return _FocalLoss(alpha, gamma, num_classes, size_average)

class _FocalLoss(torch.nn.Module):    
    def __init__(self, alpha, gamma, num_classes, size_average=True):
        """focal_loss function: -α(1-yi)**γ *ce_loss(xi,yi)   
        Args:
            alpha:  class weight (default 0.25).
                    When α is a 'list', it indicates the class-wise weights;
                    When α is a constant, 
                        if in detection task, it indicates that the class-wise weights are[α, 1-α, 1-α, ...], 
                                                the first class indicates the background
                        if in classification task, it indicates that the class-wise weights are the same
            gamma:  γ (default 2), focusing paramter smoothly adjusts the rate at which easy examples are down-weighted.
            num_classes: the number of classes
            size_average:  (default 'mean'/'sum') specify the way to compute the loss value
        """

        super(_FocalLoss,self).__init__()
        self.size_average = size_average
        if isinstance(alpha,list):
            assert len(alpha)==num_classes
            alpha /= np.sum(alpha) # setting the value in range of [0, 1]
            # print("Focal loss alpha = {}, assign specific weights for each class".format(alpha))
            self.alpha = torch.Tensor(alpha)
        else:
            assert alpha<=1   
            self.alpha = torch.zeros(num_classes)+0.00001

            # classification task
            # print("Focal loss alpha={}, the weight for each class is the same".format(alpha))
            self.alpha += alpha

            # detection task # 如果α为一个常数,则降低第一类的影响,在目标检测中背景为第一类
            # print(" --- Focal_loss alpha = {} ,将对背景类进行衰减,请在目标检测任务中使用 --- ".format(alpha))
            # self.alpha[0] += alpha
            # self.alpha[1:] += (1-alpha) # α 最终为 [ α, 1-α, 1-α, 1-α, 1-α, ...] size:[num_classes]
        self.gamma = gamma

    def forward(self, predictions, labels):
        """
        focal_loss损失计算
        Args:        
            preds:   预测类别. size:[B,N,C] or [B,C]    分别对应与检测与分类任务, B 批次, N检测框数, C类别数        
            labels:  实际类别. size:[B,N] or [B]        
        return:
            loss
        """        
        assert predictions.dim()==2 and labels.dim()==1        
        preds = predictions.view(-1,predictions.size(-1))  # num*classes
        alpha = self.alpha.to(labels.device) 

        # 这里并没有直接使用log_softmax, 因为后面会用到softmax的结果(当然你也可以使用log_softmax,然后进行exp操作)   
        preds_softmax = F.softmax(preds, dim=1)      
        preds_logsoft = torch.log(preds_softmax)

        # implement nll_loss ( crossempty = log_softmax + nll )     
        preds_softmax = preds_softmax.gather(1,labels.view(-1,1))  # num*1    
        preds_logsoft = preds_logsoft.gather(1,labels.view(-1,1))  # num*1      
        alpha = alpha.gather(0,labels.view(-1)) # num     

        # calc loss
        # torch.pow((1-preds_softmax), self.gamma) 为focal loss中 (1-pt)**γ
        # shape: num*1
        loss = -torch.mul( torch.pow((1-preds_softmax), self.gamma), preds_logsoft )  
        # α * (1-pt)**γ * ce_loss
        # shape: 
        loss = torch.mul(alpha, loss.t())    
        del preds    
        del alpha
        if self.size_average:        
            loss = loss.mean()        
        else:            
            loss = loss.sum()        
        return loss


import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.resnet import ResNet

from .build import META_ARCH_REGISTRY

__all__ = [
    'ResNet',
    'Resnet18',
    'Resnet34',
    'Resnet50',
    'Resnet101',
    'Resnet152',
    'Resnext50_32x4d',
    'Resnext101_32x8d',
    'Wide_resnet50_2',
    'Wide_resnet101_2'
]


class ResNet(nn.Module):
    # Modify attributs	    
    def __init__(self, model):
        super(ResNet, self).__init__()
        for key, val in model.__dict__.items():
            self.__dict__[key] = val
        self.g_avg_pool = nn.AdaptiveAvgPool2d(1)
        self.last_linear = self.fc

    def features(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.maxpool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        return out

    def logits(self, x):
        out = self.g_avg_pool(x)
        out = out.view(out.size(0), -1)
        out = self.last_linear(out)
        return out

    def forward(self, x):
        out = self.features(x)
        out = self.logits(out)
        return out

def generate_model(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.fc.in_features
        model.fc = nn.Linear(in_features, classes, bias=False)
    return ResNet(model)

@META_ARCH_REGISTRY.register()
def Resnet18(cfg):
    return generate_model(cfg, 'resnet18')

@META_ARCH_REGISTRY.register()
def Resnet34(cfg):
    return generate_model(cfg, 'resnet34')

@META_ARCH_REGISTRY.register()
def Resnet50(cfg):
    return generate_model(cfg, 'resnet50')

@META_ARCH_REGISTRY.register()
def Resnet101(cfg):
    return generate_model(cfg, 'resnet101')

@META_ARCH_REGISTRY.register()
def Resnet152(cfg):
    return generate_model(cfg, 'resnet152')

@META_ARCH_REGISTRY.register()
def Resnext50_32x4d(cfg):
    return generate_model(cfg, 'resnext50_32x4d')

@META_ARCH_REGISTRY.register()
def Resnext101_32x8d(cfg):
    return generate_model(cfg, 'resnext101_32x8d')

@META_ARCH_REGISTRY.register()
def Wide_resnet50_2(cfg):
    return generate_model(cfg, 'wide_resnet50_2')

@META_ARCH_REGISTRY.register()
def Wide_resnet101_2(cfg):
    return generate_model(cfg, 'wide_resnet101_2')


"""
Example template for defining a system
"""
import logging
import os
from argparse import ArgumentParser
from collections import OrderedDict

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms
from pytorch_lightning import LightningModule
from torch import optim
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
import torchvision

from torchline.data import build_data, build_sampler
from torchline.losses import build_loss_fn
from torchline.models import build_model
from torchline.utils import topk_acc, AverageMeterGroup
from .build import MODULE_REGISTRY
from .utils import generate_optimizer, generate_scheduler

__all__ = [
    'DefaultModule'
]

@MODULE_REGISTRY.register()
class DefaultModule(LightningModule):
    """
    Sample model to show how to define a template
    """

    def __init__(self, cfg):
        """
        Pass in parsed HyperOptArgumentParser to the model
        :param cfg:
        """
        # init superclass
        super(DefaultModule, self).__init__()
        self.cfg = cfg
        self.hparams = self.cfg.hparams
        self.batch_size = self.cfg.dataset.batch_size

        # if you specify an example input, the summary will show input/output for each layer
        h, w = self.cfg.input.size
        self.example_input_array = torch.rand(1, 3, h, w)

        # build model
        self.model = self.build_model(cfg)
        self.train_meters = AverageMeterGroup()
        self.valid_meters = AverageMeterGroup()

    # ---------------------
    # model SETUP
    # ---------------------
    def build_model(self, cfg):
        """
        Layout model
        :return:
        """
        return build_model(cfg)

    def build_loss_fn(self, cfg):
        """
        Layout loss_fn
        :return:
        """
        return build_loss_fn(cfg)

    def build_data(self, cfg, is_train):
        """
        Layout training dataset
        :return:
        """
        cfg.defrost()
        cfg.dataset.is_train = is_train
        cfg.freeze()
        return build_data(cfg)

    def build_sampler(self, cfg, is_train):
        """
        Layout training dataset
        :return:
        """
        cfg.defrost()
        cfg.dataset.is_train = is_train
        cfg.freeze()
        return build_sampler(cfg)

    # ---------------------
    # Hooks
    # ---------------------

    def on_train_start(self):
        ckpt_path = self.trainer.resume_from_checkpoint
        print(ckpt_path)
        if ckpt_path:
            if os.path.exists(ckpt_path):
                ckpt = torch.load(ckpt_path)
                best = ckpt['checkpoint_callback_best']
                self.log_info(f"The best result of the ckpt is {best}")
            else:
                print(f'{ckpt_path} not exists')
                raise NotImplementedError

    def on_epoch_start(self):
        if not self.cfg.trainer.progress_bar_refresh_rate:
            # print current lr
            if isinstance(self.trainer.optimizers, list):
                if len(self.trainer.optimizers) == 1:
                    optimizer = self.trainer.optimizers[0]
                    lr = optimizer.param_groups[0]["lr"]
                    self.log_info(f"lr={lr:.4e}")
                else:
                    for index, optimizer in enumerate(self.trainer.optimizers):
                        lr = optimizer.param_groups[0]["lr"]
                        name = str(optimizer).split('(')[0].strip()
                        self.log_info(f"lr of {name}_{index} is {lr:.4e} ")
            else:
                lr = self.trainer.optimizers.param_groups[0]["lr"]
                self.log_info(f"lr={lr:.4e}")

    def on_epoch_end(self):
        if not self.cfg.trainer.progress_bar_refresh_rate:
            self.log_info(f'Final Train: {self.train_meters}')
            self.log_info(f'FInal Valid: {self.valid_meters}')
            self.log_info("===========================\n")
            self.train_meters = AverageMeterGroup()
            self.valid_meters = AverageMeterGroup()

    def log_info(self, *args, **kwargs):
        if self.trainer.proc_rank == 0:
            self.trainer.logger_print.info(*args, **kwargs)

    def log_warning(self, *args, **kwargs):
        if self.trainer.proc_rank == 0:
            self.trainer.logger_print.warning(*args, **kwargs)
    # ---------------------
    # TRAINING
    # ---------------------
    
    def forward(self, x):
        """
        No special modification required for lightning, define as you normally would
        :param x:
        :return:
        
        return middle features
        features = self.model.features(x)
        logits = self.model.logits(features)
        return logits
        """
        return self.model(x)

    def loss(self, predictions, gt_labels):
        loss_fn = self.build_loss_fn(self.cfg)
        return loss_fn(predictions, gt_labels)

    def print_log(self, batch_idx, is_train, inputs, meters, save_examples=False):
        flag = batch_idx % self.cfg.trainer.log_save_interval == 0
        if not self.trainer.progress_bar_refresh_rate and flag:
            if is_train:
                _type = 'Train'
                all_step = self.trainer.num_training_batches
            else:
                _type = 'Valid'
                all_step = self.trainer.num_val_batches
            crt_epoch, crt_step = self.trainer.current_epoch, batch_idx
            all_epoch = self.trainer.max_epochs
            log_info = f"{_type} Epoch {crt_epoch}/{all_epoch} step {crt_step}/{all_step} {meters}" 
            self.log_info(log_info)

        if self.current_epoch==0 and batch_idx==0 and save_examples:
            if not os.path.exists('train_valid_samples'):
                os.makedirs('train_valid_samples')
            for i, img in enumerate(inputs[:5]):
                torchvision.transforms.ToPILImage()(img.cpu()).save(f'./train_valid_samples/{_type}_img{i}.jpg')

    def training_step(self, batch, batch_idx):
        """
        Lightning calls this inside the training loop
        :param batch:
        :return:
        """

        # forward pass
        inputs, gt_labels = batch
        predictions = self.forward(inputs)

        # calculate loss
        loss_val = self.loss(predictions, gt_labels)

        # acc
        acc_results = topk_acc(predictions, gt_labels, self.cfg.topk)
        tqdm_dict = {}

        if self.on_gpu:
            acc_results = [torch.tensor(x).to(loss_val.device.index) for x in acc_results]

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_val = loss_val.unsqueeze(0)
            acc_results = [x.unsqueeze(0) for x in acc_results]

        tqdm_dict['train_loss'] = loss_val
        for i, k in enumerate(self.cfg.topk):
            tqdm_dict[f'train_acc_{k}'] = acc_results[i]

        output = OrderedDict({
            'loss': loss_val,
            'progress_bar': tqdm_dict,
            'log': tqdm_dict
        })

        self.train_meters.update({key: val.item() for key, val in tqdm_dict.items()})
        self.print_log(batch_idx, True, inputs, self.train_meters)

        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_step(self, batch, batch_idx):
        """
        Lightning calls this inside the validation loop
        :param batch:
        :return:
        """
        inputs, gt_labels = batch
        predictions = self.forward(inputs)

        loss_val = self.loss(predictions, gt_labels)

        # acc
        val_acc_1, val_acc_k = topk_acc(predictions, gt_labels, self.cfg.topk)

        if self.on_gpu:
            val_acc_1 = val_acc_1.cuda(loss_val.device.index)
            val_acc_k = val_acc_k.cuda(loss_val.device.index)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_val = loss_val.unsqueeze(0)
            val_acc_1 = val_acc_1.unsqueeze(0)
            val_acc_k = val_acc_k.unsqueeze(0)
        
        output = OrderedDict({
            'valid_loss': loss_val,
            'valid_acc_1': val_acc_1,
            f'valid_acc_{self.cfg.topk[-1]}': val_acc_k,
        })
        tqdm_dict = {k: v for k, v in dict(output).items()}
        self.valid_meters.update({key: val.item() for key, val in tqdm_dict.items()})
        self.print_log(batch_idx, False, inputs, self.valid_meters)

        # can also return just a scalar instead of a dict (return loss_val)
        return output

    def validation_end(self, outputs):
        """
        Called at the end of validation to aggregate outputs
        :param outputs: list of individual outputs of each validation step
        :return:
        """
        # if returned a scalar from validation_step, outputs is a list of tensor scalars
        # we return just the average in this case (if we want)
        # return torch.stack(outputs).mean()

        tqdm_dict = {key: val.avg for key, val in self.valid_meters.meters.items()}
        result = {'progress_bar': tqdm_dict, 'log': tqdm_dict, 'valid_loss': self.valid_meters.meters['valid_loss'].avg}
        return result

    def test_step(self, batch, batch_idx):
        return self.validation_step(batch, batch_idx)
    
    def test_end(self, outputs):
        return self.validation_end(outputs)

    # ---------------------
    # TRAINING SETUP
    # ---------------------
    @classmethod
    def parse_cfg_for_scheduler(cls, cfg, scheduler_name):
        if scheduler_name.lower() == 'CosineAnnealingLR'.lower():
            params = {'T_max': cfg.optim.scheduler.t_max}
        elif scheduler_name.lower() == 'CosineAnnealingWarmRestarts'.lower():
            params = {'T_0': cfg.optim.scheduler.t_0, 'T_mult': cfg.optim.scheduler.t_mul}
        elif scheduler_name.lower() == 'StepLR'.lower():
            params = {'step_size': cfg.optim.scheduler.step_size, 'gamma': cfg.optim.scheduler.gamma}
        elif scheduler_name.lower() == 'MultiStepLR'.lower():
            params = {'milestones': cfg.optim.scheduler.milestones, 'gamma': cfg.optim.scheduler.gamma}
        elif scheduler_name.lower() == 'ReduceLROnPlateau'.lower():
            params = {'mode': cfg.optim.scheduler.mode, 'patience': cfg.optim.scheduler.patience, 
                      'verbose': cfg.optim.scheduler.verbose, 'factor': cfg.optim.scheduler.gamma}
        else:
            print(f"{scheduler_name} not implemented")
            raise NotImplementedError
        return params

    def configure_optimizers(self):
        """
        return whatever optimizers we want here
        :return: list of optimizers
        """
        optim_name = self.cfg.optim.name
        momentum = self.cfg.optim.momentum
        weight_decay = self.cfg.optim.weight_decay
        lr = self.cfg.optim.base_lr
        optimizer = generate_optimizer(self.model, optim_name, lr, momentum, weight_decay)
        scheduler_params = self.parse_cfg_for_scheduler(self.cfg, self.cfg.optim.scheduler.name)
        scheduler = generate_scheduler(optimizer, self.cfg.optim.scheduler.name, **scheduler_params)
        return [optimizer], [scheduler]

    def __dataloader(self, is_train):
        # init data generators
        dataset = self.build_data(self.cfg, is_train)

        # when using multi-node (ddp) we need to add the  datasampler
        train_sampler = self.build_sampler(self.cfg, is_train)
        batch_size = self.batch_size

        if self.use_ddp:
            train_sampler = DistributedSampler(dataset)

        should_shuffle = train_sampler is None
        loader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=should_shuffle if is_train else False,
            sampler=train_sampler,
            num_workers=self.cfg.dataloader.num_workers
        )

        return loader

    @pl.data_loader
    def train_dataloader(self):
        logging.info('training data loader called')
        return self.__dataloader(is_train=True)

    @pl.data_loader
    def val_dataloader(self):
        logging.info('val data loader called')
        return self.__dataloader(is_train=False)

    @pl.data_loader
    def test_dataloader(self):
        logging.info('test data loader called')
        return self.__dataloader(is_train=False)



import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.mnasnet import MNASNet

from .build import META_ARCH_REGISTRY

__all__ = [
    'MNASNet',
    'MNASNet0_5',
    'MNASNet0_75',
    'MNASNet1_0',
    'MNASNet1_3',
]

class MNASNet(nn.Module):
    # Modify attributs	    
    def __init__(self, model):
        super(MNASNet, self).__init__()
        for key, val in model.__dict__.items():
            self.__dict__[key] = val
        self.stem   = model.layers[:8]
        self.layer1 = model.layers[8]
        self.layer2 = model.layers[9]
        self.layer3 = model.layers[10]
        self.layer4 = model.layers[11]
        self.layer5 = model.layers[12]
        self.layer6 = model.layers[13]
        self.layer7 = model.layers[14:]
        self.g_avg_pool = nn.AdaptiveAvgPool2d(1)
        self.last_linear = model.classifier

    def features(self, x): # b*3*64*64
        out = self.stem(x)   # b*16*32*32
        out = self.layer1(out)     # b*16*16*16
        out = self.layer2(out)     # b*24*8*8
        out = self.layer3(out)     # b*40*4*4
        out = self.layer4(out)     # b*48*4*4
        out = self.layer5(out)     # b*96*2*2
        out = self.layer6(out)     # b*160*2*2
        out = self.layer7(out)     # b*1280*2*2
        return out

    def logits(self, x):
        out = x.mean([2, 3])
        out = self.last_linear(out)
        return out

    def forward(self, x):
        out = self.features(x)
        out = self.logits(out)
        return out

def generate_model(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    if 'dropout' in cfg.model:
        dropout = cfg.model.dropout
    else:
        dropout = 0.2
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.classifier[1].in_features
        model.classifier = nn.Sequential(
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(in_features, classes, bias=False))
    return MNASNet(model)

@META_ARCH_REGISTRY.register()
def MNASNet0_5(cfg):
    return generate_model(cfg, 'mnasnet0_5')
    
@META_ARCH_REGISTRY.register()
def MNASNet0_75(cfg):
    return generate_model(cfg, 'mnasnet0_75')
    
@META_ARCH_REGISTRY.register()
def MNASNet1_0(cfg):
    return generate_model(cfg, 'mnasnet1_0')
    
@META_ARCH_REGISTRY.register()
def MNASNet1_3(cfg):
    return generate_model(cfg, 'mnasnet1_3')


from torchline.utils import Registry

TRAINER_REGISTRY = Registry("TRAINER")
TRAINER_REGISTRY.__doc__ = """
Registry for meta-architectures, i.e. the whole model.

The registered object will be called with `obj(cfg)`
and expected to return a `nn.Module` object.
"""

__all__ = [
    'TRAINER_REGISTRY',
    'build_trainer',
]

def build_trainer(cfg, hparams):
    """
    Built the whole trainer, defined by `cfg.trainer.name`.
    """
    name = cfg.trainer.name
    return TRAINER_REGISTRY.get(name)(cfg, hparams)
    


#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# refernce: https://github.com/facebookresearch/fvcore/blob/51695ec2984ae7b4afae27a14d02986ef3f8afd8/fvcore/common/registry.py

from typing import Dict, Optional

__all__ = [
    'Registry'
]

class Registry(object):
    '''
    The registry that provides name -> object mapping, to support third-party
    users' custom modules.
    To create a registry (e.g. a backbone registry):
    .. code-block:: python
        BACKBONE_REGISTRY = Registry('BACKBONE')
    To register an object:
    .. code-block:: python
        @BACKBONE_REGISTRY.register()
        class MyBackbone():
            ...
    Or:
    .. code-block:: python
        BACKBONE_REGISTRY.register(MyBackbone)
    '''

    def __init__(self, name: str) -> None:
        """
        Args:
            name (str): the name of this registry
        """
        self._name: str = name
        self._obj_map: Dict[str, object] = {}

    def _do_register(self, name: str, obj: object) -> None:
        assert (
            name not in self._obj_map
        ), "An object named '{}' was already registered in '{}' registry!".format(
            name, self._name
        )
        self._obj_map[name] = obj

    def register(self, obj: object = None) -> Optional[object]:
        '''
        Register the given object under the the name `obj.__name__`.
        Can be used as either a decorator or not. See docstring of this class for usage.
        '''
        if obj is None:
            # used as a decorator
            def deco(func_or_class: object) -> object:
                name = func_or_class.__name__  # pyre-ignore
                self._do_register(name, func_or_class)
                return func_or_class

            return deco

        # used as a function call
        name = obj.__name__  # pyre-ignore
        self._do_register(name, obj)

    def get(self, name: str) -> object:
        ret = self._obj_map.get(name)
        if ret is None:
            raise KeyError(
                "No object named '{}' found in '{}' registry!".format(
                    name, self._name
                )
            )
        return ret



#coding=utf-8
import torch
import numpy as np
import math
import random
import torchvision
from torchvision import transforms
from . import autoaugment
from torchline.utils import Registry, Logger

TRANSFORMS_REGISTRY = Registry('transforms')
TRANSFORMS_REGISTRY.__doc__ = """
Registry for data transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

LABEL_TRANSFORMS_REGISTRY = Registry('label_transforms')
LABEL_TRANSFORMS_REGISTRY.__doc__ = """
Registry for label transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

__all__ = [
    'build_transforms',
    'build_label_transforms',
    'TRANSFORMS_REGISTRY',
    'LABEL_TRANSFORMS_REGISTRY',
    'DefaultTransforms',
    '_DefaultTransforms',
    'BaseTransforms'
]


def build_transforms(cfg):
    """
    Built the transforms, defined by `cfg.transforms.name`.
    """
    name = cfg.transforms.name
    return TRANSFORMS_REGISTRY.get(name)(cfg)
    
def build_label_transforms(cfg):
    """
    Built the label transforms, defined by `cfg.label_transforms.name`.
    """
    name = cfg.label_transforms.name
    if name == 'default':
        return None
    return LABEL_TRANSFORMS_REGISTRY.get(name)(cfg)

class BaseTransforms(object):
    def __init__(self, is_train, log_name):
        self.logger_print = Logger(__name__, log_name).getlogger()
        self.is_train = is_train

    def get_transform(self):
        if not self.is_train: 
            self.logger_print.info('Generating validation transform ...')
            transform = self.valid_transform
            self.logger_print.info(f'Valid transform={transform}')
        else:
            self.logger_print.info('Generating training transform ...')
            transform = self.train_transform
            self.logger_print.info(f'Train transform={transform}')
        return transform

    @property
    def valid_transform(self):
        raise NotImplementedError

    @property
    def train_transform(self):
        raise NotImplementedError


@TRANSFORMS_REGISTRY.register()
def DefaultTransforms(cfg):
    is_train = cfg.dataset.is_train
    log_name = cfg.log.name
    mean = cfg.transforms.tensor.normalization.mean
    std = cfg.transforms.tensor.normalization.std
    img_size = cfg.input.size
    
    aug_imagenet = cfg.transforms.img.aug_imagenet
    aug_cifar = cfg.transforms.img.aug_cifar
    random_resized_crop = cfg.transforms.img.random_resized_crop
    resize = cfg.transforms.img.resize
    random_crop = cfg.transforms.img.random_crop
    center_crop = cfg.transforms.img.center_crop
    random_horizontal_flip = cfg.transforms.img.random_horizontal_flip
    random_vertical_flip = cfg.transforms.img.random_vertical_flip
    random_rotation = cfg.transforms.img.random_rotation
    color_jitter = cfg.transforms.img.color_jitter
    return _DefaultTransforms(is_train, log_name, img_size,
                 aug_imagenet, aug_cifar,
                 random_resized_crop,
                 resize,
                 random_crop,
                 center_crop,
                 random_horizontal_flip,
                 random_vertical_flip,
                 random_rotation,
                 color_jitter,
                 mean, std)

class _DefaultTransforms(BaseTransforms):
    def __init__(self, is_train, log_name, img_size,
                 aug_imagenet=False, aug_cifar=False,
                 random_resized_crop={'enable':0},
                 resize={'enable':1},
                 random_crop={'enable':0, 'padding':0},
                 center_crop={'enable':0},
                 random_horizontal_flip={'enbale':0, 'p':0.5},
                 random_vertical_flip={'enbale':0, 'p':0.5},
                 random_rotation={'enbale':0, 'degrees':15},
                 color_jitter={'enable':0,'brightness':0.1, 'contrast':0.1, 'saturation':0.1, 'hue':0.1},
                 mean=[0.5,0.5,0.5], std=[0.5,0.5,0.5], *args, **kwargs):
        super(_DefaultTransforms, self).__init__(is_train, log_name)
        self.is_train = is_train
        self.mean = mean
        self.std = std
        self.img_size = img_size
        self.min_edge_size = min(self.img_size)
        self.normalize = transforms.Normalize(self.mean, self.std)
        self.aug_imagenet = aug_imagenet
        self.aug_cifar = aug_cifar
        self.random_resized_crop = random_resized_crop
        self.resize = resize
        self.random_crop = random_crop
        self.center_crop = center_crop
        self.random_horizontal_flip = random_horizontal_flip
        self.random_vertical_flip = random_vertical_flip
        self.random_rotation = random_rotation
        self.color_jitter = color_jitter
        self.transform = self.get_transform()

    @property
    def valid_transform(self):
        transform = transforms.Compose([
            transforms.Resize(self.img_size),
            transforms.ToTensor(),
            self.normalize
        ])
        return transform

    @property
    def train_transform(self):
        # aug_imagenet
        if self.aug_imagenet:
            self.logger_print.info('Using imagenet augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.img_size),
                autoaugment.ImageNetPolicy(),
                transforms.ToTensor(),
                self.normalize
            ])
        # aug cifar
        elif self.aug_cifar:
            self.logger_print.info('Using cifar augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.img_size),
                autoaugment.CIFAR10Policy(),
                transforms.ToTensor(),
                self.normalize
            ])
        # customized transformations
        else:
            transform = self.read_transform_from_cfg()
        return transform

    def read_transform_from_cfg(self):
        transform_list = []
        self.check_conflict_options()

        # resize and crop opertaion
        if self.random_resized_crop['enable']:
            transform_list.append(transforms.RandomResizedCrop(self.img_size))
        elif self.resize['enable']:
            transform_list.append(transforms.Resize(self.img_size))
        if self.random_crop['enable']:
            transform_list.append(transforms.RandomCrop(self.min_edge_size, padding=self.random_crop['padding']))
        elif self.center_crop['enable']:
            transform_list.append(transforms.CenterCrop(self.min_edge_size))

        # ColorJitter
        if self.color_jitter['enable']:
            params = {key: self.color_jitter[key] for key in self.color_jitter 
                            if key != 'enable'}
            transform_list.append(transforms.ColorJitter(**params))

        # horizontal flip
        if self.random_horizontal_flip['enable']:
            p = self.random_horizontal_flip['p']
            transform_list.append(transforms.RandomHorizontalFlip(p))
        
        # vertical flip
        if self.random_vertical_flip['enable']:
            p = self.random_vertical_flip['p']
            transform_list.append(transforms.RandomVerticalFlip(p))

        # rotation
        if self.random_rotation['enable']:
            degrees = self.random_rotation['degrees']
            transform_list.append(transforms.RandomRotation(degrees))
        transform_list.append(transforms.ToTensor())
        transform_list.append(self.normalize)
        transform_list = transforms.Compose(transform_list)
        assert len(transform_list.transforms) > 0, "You must apply transformations"
        return transform_list

    def check_conflict_options(self):
        count = self.random_resized_crop['enable'] + \
                self.resize['enable']
        assert count <= 1, 'You can only use one resize transform operation'

        count = self.random_crop['enable'] + \
                self.center_crop['enable']
        assert count <= 1, 'You can only use one crop transform operation'